@@ -3,10 +3,13 @@ use core::marker;
 use core::mem;
 use core::str;
 
-use musli::de::PackDecoder;
-use musli::de::SequenceDecoder;
-use musli::de::{Decoder, PairDecoder, PairsDecoder, ValueVisitor, VariantDecoder};
+use musli::de::{
+    Decoder, LengthHint, NumberHint, NumberVisitor, PackDecoder, PairDecoder, PairsDecoder,
+    SequenceDecoder, TypeHint, ValueVisitor, VariantDecoder,
+};
 use musli::error::Error;
+#[cfg(feature = "std")]
+use musli::mode::Mode;
 use musli::never::Never;
 
 use crate::reader::integer::{Signed, Unsigned};
@@ -106,13 +109,16 @@ where
     P: Parser<'de>,
 {
     type Error = ParseError;
-    type Pack = Never<Self>;
+    #[cfg(not(feature = "std"))]
+    type Buffer = Never<Self::Error>;
+    #[cfg(feature = "std")]
+    type Buffer = musli_value::AsValueDecoder<Self::Error>;
+    type Pack = Never<Self::Error>;
     type Sequence = JsonSequenceDecoder<'a, P>;
     type Tuple = JsonSequenceDecoder<'a, P>;
     type Map = JsonObjectDecoder<'a, P>;
     type Some = JsonDecoder<'a, P>;
     type Struct = JsonObjectDecoder<'a, P>;
-    type TupleStruct = JsonObjectDecoder<'a, P>;
     type Variant = JsonVariantDecoder<'a, P>;
 
     #[inline]
@@ -120,6 +126,31 @@ where
         write!(f, "value that can be decoded from JSON")
     }
 
+    #[inline]
+    fn type_hint(&mut self) -> Result<TypeHint, Self::Error> {
+        Ok(match self.parser.peek()? {
+            Token::OpenBrace => TypeHint::Map(LengthHint::Any),
+            Token::OpenBracket => TypeHint::Sequence(LengthHint::Any),
+            Token::String => TypeHint::String(LengthHint::Any),
+            Token::Number => TypeHint::Number(NumberHint::Any),
+            Token::Null => TypeHint::Unit,
+            Token::True => TypeHint::Bool,
+            Token::False => TypeHint::Bool,
+            _ => TypeHint::Any,
+        })
+    }
+
+    #[cfg(feature = "std")]
+    #[inline]
+    fn decode_buffer<M>(self) -> Result<Self::Buffer, Self::Error>
+    where
+        M: Mode,
+    {
+        use musli::de::Decode;
+        let value: musli_value::Value = Decode::<M>::decode(self)?;
+        Ok(value.into_value_decoder())
+    }
+
     #[inline]
     fn decode_unit(self) -> Result<(), Self::Error> {
         self.skip_any()
@@ -143,64 +174,86 @@ where
         }
     }
 
+    #[inline]
+    fn decode_char(mut self) -> Result<char, Self::Error> {
+        let start = self.parser.pos();
+
+        let string = match self.parser.parse_string(self.scratch, true)? {
+            StringReference::Borrowed(string) => string,
+            StringReference::Scratch(string) => string,
+        };
+
+        let mut it = string.chars();
+        let first = it.next();
+
+        match (first, it.next()) {
+            (Some(c), None) => Ok(c),
+            _ => Err(ParseError::spanned(
+                start,
+                self.parser.pos(),
+                ParseErrorKind::CharEmptyString,
+            )),
+        }
+    }
+
     #[inline]
     fn decode_u8(mut self) -> Result<u8, Self::Error> {
-        integer::decode_unsigned(&mut self.parser)
+        integer::parse_unsigned(&mut self.parser)
     }
 
     #[inline]
     fn decode_u16(mut self) -> Result<u16, Self::Error> {
-        integer::decode_unsigned(&mut self.parser)
+        integer::parse_unsigned(&mut self.parser)
     }
 
     #[inline]
     fn decode_u32(mut self) -> Result<u32, Self::Error> {
-        integer::decode_unsigned(&mut self.parser)
+        integer::parse_unsigned(&mut self.parser)
     }
 
     #[inline]
     fn decode_u64(mut self) -> Result<u64, Self::Error> {
-        integer::decode_unsigned(&mut self.parser)
+        integer::parse_unsigned(&mut self.parser)
     }
 
     #[inline]
     fn decode_u128(mut self) -> Result<u128, Self::Error> {
-        integer::decode_unsigned(&mut self.parser)
+        integer::parse_unsigned(&mut self.parser)
     }
 
     #[inline]
     fn decode_i8(mut self) -> Result<i8, Self::Error> {
-        integer::decode_signed(&mut self.parser)
+        integer::parse_signed(&mut self.parser)
     }
 
     #[inline]
     fn decode_i16(mut self) -> Result<i16, Self::Error> {
-        integer::decode_signed(&mut self.parser)
+        integer::parse_signed(&mut self.parser)
     }
 
     #[inline]
     fn decode_i32(mut self) -> Result<i32, Self::Error> {
-        integer::decode_signed(&mut self.parser)
+        integer::parse_signed(&mut self.parser)
     }
 
     #[inline]
     fn decode_i64(mut self) -> Result<i64, Self::Error> {
-        integer::decode_signed(&mut self.parser)
+        integer::parse_signed(&mut self.parser)
     }
 
     #[inline]
     fn decode_i128(mut self) -> Result<i128, Self::Error> {
-        integer::decode_signed(&mut self.parser)
+        integer::parse_signed(&mut self.parser)
     }
 
     #[inline]
     fn decode_usize(mut self) -> Result<usize, Self::Error> {
-        integer::decode_unsigned(&mut self.parser)
+        integer::parse_unsigned(&mut self.parser)
     }
 
     #[inline]
     fn decode_isize(mut self) -> Result<isize, Self::Error> {
-        integer::decode_signed(&mut self.parser)
+        integer::parse_signed(&mut self.parser)
     }
 
     #[inline]
@@ -213,6 +266,14 @@ where
         self.parser.parse_f64()
     }
 
+    #[inline]
+    fn decode_number<V>(mut self, visitor: V) -> Result<V::Ok, Self::Error>
+    where
+        V: NumberVisitor<Error = Self::Error>,
+    {
+        self.parser.parse_number(visitor)
+    }
+
     #[inline]
     fn decode_option(mut self) -> Result<Option<Self::Some>, Self::Error> {
         if self.parser.peek()?.is_null() {
@@ -223,34 +284,29 @@ where
         }
     }
 
+    #[inline]
+    fn decode_bytes<V>(self, visitor: V) -> Result<V::Ok, V::Error>
+    where
+        V: ValueVisitor<'de, Target = [u8], Error = Self::Error>,
+    {
+        let mut seq = self.decode_sequence()?;
+        let mut bytes = Vec::with_capacity(seq.size_hint().unwrap_or_default());
+
+        while let Some(item) = SequenceDecoder::next(&mut seq)? {
+            bytes.push(item.decode_u8()?);
+        }
+
+        visitor.visit_owned(bytes)
+    }
+
     #[inline]
     fn decode_string<V>(mut self, visitor: V) -> Result<V::Ok, V::Error>
     where
         V: ValueVisitor<'de, Target = str, Error = Self::Error>,
     {
-        let actual = self.parser.peek()?;
-
-        if !matches!(actual, Token::String) {
-            return Err(V::Error::message(format_args!(
-                "expected string, but was {actual}"
-            )));
-        }
-
-        self.parser.skip(1)?;
-
         match self.parser.parse_string(self.scratch, true)? {
-            StringReference::Borrowed(borrowed) => {
-                // SAFETY: safety is guaranteed by the implementation of
-                // `parse_string`.
-                let string = unsafe { str::from_utf8_unchecked(borrowed) };
-                visitor.visit_borrowed(string)
-            }
-            StringReference::Scratch(string) => {
-                // SAFETY: safety is guaranteed by the implementation of
-                // `parse_string`.
-                let string = unsafe { str::from_utf8_unchecked(string) };
-                visitor.visit_any(string)
-            }
+            StringReference::Borrowed(borrowed) => visitor.visit_borrowed(borrowed),
+            StringReference::Scratch(string) => visitor.visit_any(string),
         }
     }
 
@@ -274,16 +330,6 @@ where
         JsonObjectDecoder::new(self.scratch, Some(len), self.parser)
     }
 
-    #[inline]
-    fn decode_tuple_struct(self, len: usize) -> Result<Self::Struct, Self::Error> {
-        JsonObjectDecoder::new(self.scratch, Some(len), self.parser)
-    }
-
-    #[inline]
-    fn decode_unit_struct(self) -> Result<(), Self::Error> {
-        self.skip_any()
-    }
-
     #[inline]
     fn decode_variant(self) -> Result<Self::Variant, Self::Error> {
         JsonVariantDecoder::new(self.scratch, self.parser)
@@ -300,6 +346,7 @@ impl<'de, 'a, P> JsonKeyDecoder<'a, P>
 where
     P: Parser<'de>,
 {
+    #[inline]
     fn skip_any(self) -> Result<(), ParseError> {
         JsonDecoder::new(self.scratch, self.parser).skip_any()
     }
@@ -320,19 +367,9 @@ where
     where
         V: ValueVisitor<'de, Target = [u8], Error = ParseError>,
     {
-        let actual = self.parser.peek()?;
-
-        if !matches!(actual, Token::String) {
-            return Err(V::Error::message(format_args!(
-                "expected string, but was {actual}"
-            )));
-        }
-
-        self.parser.skip(1)?;
-
         match self.parser.parse_string(self.scratch, true)? {
-            StringReference::Borrowed(bytes) => visitor.visit_borrowed(bytes),
-            StringReference::Scratch(bytes) => visitor.visit_any(bytes),
+            StringReference::Borrowed(string) => visitor.visit_borrowed(string.as_bytes()),
+            StringReference::Scratch(string) => visitor.visit_any(string.as_bytes()),
         }
     }
 }
@@ -364,7 +401,7 @@ where
 
     #[inline]
     fn visit_any(self, bytes: &Self::Target) -> Result<Self::Ok, Self::Error> {
-        integer::decode_unsigned(&mut &mut SliceParser::new(bytes))
+        integer::parse_unsigned(&mut &mut SliceParser::new(bytes))
     }
 }
 
@@ -395,7 +432,7 @@ where
 
     #[inline]
     fn visit_any(self, bytes: &Self::Target) -> Result<Self::Ok, Self::Error> {
-        integer::decode_signed(&mut SliceParser::new(bytes))
+        integer::parse_signed(&mut SliceParser::new(bytes))
     }
 }
 
@@ -404,20 +441,25 @@ where
     P: Parser<'de>,
 {
     type Error = ParseError;
-    type Pack = Never<Self>;
-    type Sequence = Never<Self>;
-    type Tuple = Never<Self>;
-    type Map = Never<Self>;
-    type Some = Never<Self>;
+    type Buffer = Never<Self::Error>;
+    type Pack = Never<Self::Error>;
+    type Sequence = Never<Self::Error>;
+    type Tuple = Never<Self::Error>;
+    type Map = Never<Self::Error>;
+    type Some = Never<Self::Error>;
     type Struct = JsonObjectDecoder<'a, P>;
-    type TupleStruct = Never<Self>;
-    type Variant = Never<Self>;
+    type Variant = Never<Self::Error>;
 
     #[inline]
     fn expecting(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
         write!(f, "value that can be decoded from a object key")
     }
 
+    #[inline]
+    fn type_hint(&mut self) -> Result<TypeHint, Self::Error> {
+        JsonDecoder::new(self.scratch, &mut self.parser).type_hint()
+    }
+
     #[inline]
     fn decode_u8(self) -> Result<u8, Self::Error> {
         self.decode_escaped_bytes(KeyUnsignedVisitor::new())
@@ -569,6 +611,11 @@ where
             }
         }
     }
+
+    #[inline]
+    fn end(self) -> Result<(), Self::Error> {
+        Ok(())
+    }
 }
 
 pub struct JsonObjectPairDecoder<'a, P> {
@@ -638,6 +685,7 @@ pub struct JsonSequenceDecoder<'a, P> {
     len: Option<usize>,
     first: bool,
     parser: P,
+    terminated: bool,
 }
 
 impl<'de, 'a, P> JsonSequenceDecoder<'a, P>
@@ -668,6 +716,7 @@ where
             len,
             first: true,
             parser,
+            terminated: false,
         })
     }
 }
@@ -682,10 +731,12 @@ where
     where
         Self: 'this;
 
+    #[inline]
     fn size_hint(&self) -> Option<usize> {
         self.len
     }
 
+    #[inline]
     fn next(&mut self) -> Result<Option<Self::Decoder<'_>>, Self::Error> {
         let first = mem::take(&mut self.first);
 
@@ -705,6 +756,7 @@ where
                 }
                 Token::CloseBracket => {
                     self.parser.skip(1)?;
+                    self.terminated = true;
                     return Ok(None);
                 }
                 _ => {
@@ -715,6 +767,24 @@ where
             }
         }
     }
+
+    #[inline]
+    fn end(mut self) -> Result<(), Self::Error> {
+        if !self.terminated {
+            let actual = self.parser.peek()?;
+
+            if !matches!(actual, Token::CloseBracket) {
+                return Err(ParseError::at(
+                    self.parser.pos(),
+                    ParseErrorKind::ExpectedCloseBracket(actual),
+                ));
+            }
+
+            self.terminated = true;
+        }
+
+        Ok(())
+    }
 }
 
 impl<'de, 'a, P> PackDecoder<'de> for JsonSequenceDecoder<'a, P>
@@ -727,6 +797,7 @@ where
     where
         Self: 'this;
 
+    #[inline]
     fn next(&mut self) -> Result<Self::Decoder<'_>, Self::Error> {
         let first = mem::take(&mut self.first);
 
@@ -743,6 +814,7 @@ where
                 }
                 Token::CloseBracket => {
                     self.parser.skip(1)?;
+                    self.terminated = true;
 
                     return Err(Self::Error::message(format_args!(
                         "encountered short array, but found {token}"
@@ -756,6 +828,24 @@ where
             }
         }
     }
+
+    #[inline]
+    fn end(mut self) -> Result<(), Self::Error> {
+        if !self.terminated {
+            let actual = self.parser.peek()?;
+
+            if !matches!(actual, Token::CloseBracket) {
+                return Err(ParseError::at(
+                    self.parser.pos(),
+                    ParseErrorKind::ExpectedCloseBracket(actual),
+                ));
+            }
+
+            self.terminated = true;
+        }
+
+        Ok(())
+    }
 }
 
 pub struct JsonVariantDecoder<'a, P> {
